from __future__ import annotations

from tortoise import fields
from tortoise.models import Model


class User(Model):
    id = fields.IntField(pk=True)
    email = fields.CharField(max_length=255, unique=True)
    is_active = fields.BooleanField(default=True)
    is_superuser = fields.BooleanField(default=False)
    password = fields.CharField(max_length=255, null=True)





from __future__ import annotations

from collections.abc import Callable
from importlib import import_module

from fastapi import FastAPI

try:
    from tortoise.models import Model
except Exception:  # pragma: no cover - only if tortoise missing in env
    class Model:  # type: ignore
        pass

from .settings import Settings


class AdminSite:
    """
    Minimal admin facade. Encapsulates admin mounting and optional model registration.
    Keeps fastapi-admin & Tortoise as internal details.
    """

    def __init__(self, title: str = "Admin") -> None:
        self.title = title
        self._mounted = False
        self._models: list[type[Model]] = []

    def mount(self, app: FastAPI, settings: Settings, path: str | None = None) -> None:
        mount_path = path or settings.admin_path
        try:
            from fastapi_admin.app import app as admin_app  # type: ignore  # noqa: PLC0415

            app.mount(mount_path, admin_app)
            self._mounted = True
        except Exception:
            placeholder = FastAPI(title=f"{self.title} placeholder")

            @placeholder.get("/")
            def _placeholder() -> dict[str, str]:
                return {"admin": "not-configured"}

            app.mount(mount_path, placeholder)
            self._mounted = True

    def register_model(self, model: type[Model]) -> None:
        self._models.append(model)
        # Note: fastapi-admin resource registration can be integrated here in future


def try_call_admin_hooks(app: FastAPI, settings: Settings) -> None:
    """
    If an installed app defines `admin.init_admin(app, settings)`, call it.
    Allows projects to wire AdminSite without exposing fastapi-admin directly.
    """
    for app_name in settings.installed_apps:
        try:
            module = import_module(f"{app_name}.admin")
        except ModuleNotFoundError:
            continue
        hook: Callable[[FastAPI, Settings], None] | None = getattr(module, "init_admin", None)
        if callable(hook):
            try:
                hook(app, settings)
            except Exception:
                continue





from __future__ import annotations

import contextlib
import os
import re
import shutil
import subprocess
from pathlib import Path

import typer
from passlib.context import CryptContext
from tortoise.exceptions import DoesNotExist

from fast_django.utils import discover_models_modules

app = typer.Typer(help="fast-django CLI")


def _write_aerich_ini(location: Path) -> None:
    Path("aerich.ini").write_text(
        f"""
[aerich]
tortoise_orm = orm_config.ORM
location = {location!s}
src_folder = .
""",
        encoding="utf-8",
    )


def copy_tree(src: Path, dst: Path) -> None:
    if not dst.exists():
        dst.mkdir(parents=True)
    for root, _dirs, files in os.walk(src):
        rel = Path(root).relative_to(src)
        target_root = dst / rel
        target_root.mkdir(parents=True, exist_ok=True)
        for f in files:
            shutil.copy2(Path(root) / f, target_root / f)


@app.command()
def startproject(name: str) -> None:
    base = Path.cwd() / name
    if base.exists():
        typer.echo(f"Directory {name} already exists", err=True)
        raise typer.Exit(1)

    template = Path(__file__).parent.parent / "scaffolding" / "project"
    copy_tree(template, base)

    # rename package placeholder
    (base / "project_name").rename(base / name)

    # replace placeholders in text files
    pattern = re.compile(r"project_name")
    for path in base.rglob("*.py"):
        content = path.read_text(encoding="utf-8")
        new_content = pattern.sub(name, content)
        path.write_text(new_content, encoding="utf-8")

    # generate aerich config
    (base / "orm_config.py").write_text(
        (
            "ORM = {\n"
            "    'connections': {'default': 'sqlite://db.sqlite3'},\n"
            f"    'apps': {{'models': {{'models': ['{name}.models', 'aerich.models'], 'default_connection': 'default'}}}}\n"
            "}"
        ),
        encoding="utf-8",
    )

    typer.echo(f"Created project {name}")


@app.command()
def startapp(name: str) -> None:
    base = Path.cwd() / name
    if base.exists():
        typer.echo(f"App {name} already exists", err=True)
        raise typer.Exit(1)
    template = Path(__file__).parent.parent / "scaffolding" / "app"
    copy_tree(template, base)
    # minimal migrations folder
    (base / "migrations").mkdir(parents=True, exist_ok=True)
    (base / "migrations" / "__init__.py").write_text("\n", encoding="utf-8")
    typer.echo(f"Created app {name}")


@app.command()
def runserver(host: str = "127.0.0.1", port: int = 8000, reload: bool = True) -> None:
    # try to infer current project module as <cwd>/<name>/asgi.py
    asgi_target = None
    cwd = Path.cwd()
    for candidate in cwd.iterdir():
        if candidate.is_dir() and (candidate / "asgi.py").exists():
            asgi_target = f"{candidate.name}.asgi:app"
            break
    if asgi_target is None:
        asgi_target = "app:app"  # fallback
    cmd = ["uvicorn", asgi_target, "--host", host, "--port", str(port)]
    if reload:
        cmd.append("--reload")
    # In tests, avoid blocking server start
    if os.getenv("FD_TEST_MODE") == "1" or os.getenv("PYTEST_CURRENT_TEST"):
        typer.echo("runserver (test mode): " + " ".join(cmd))
        raise typer.Exit(0)
    subprocess.run(cmd, check=False)


@app.command()
def makemigrations(
    app: str | None = typer.Option(None, "--app", "--app-name", help="Limit migrations to given app directory"),
) -> None:
    # Ensure we're in a generated project directory
    if not (Path("manage.py").exists() or Path("pyproject.toml").exists()):
        typer.echo("Run inside a project directory.", err=True)
        raise typer.Exit(1)
    if not Path("aerich.ini").exists():
        # generate orm_config from discovered apps
        models_list = discover_models_modules(Path.cwd() if app is None else Path.cwd() / app)
        Path("orm_config.py").write_text(
            (
                "ORM = {\n"
                "    'connections': {'default': 'sqlite://db.sqlite3'},\n"
                f"    'apps': {{'models': {{'models': {models_list!r}, 'default_connection': 'default'}}}}\n"
                "}"
            ),
            encoding="utf-8",
        )
        # location per app if provided
        location = Path("migrations") if app is None else Path(app) / "migrations"
        location.mkdir(parents=True, exist_ok=True)
        _write_aerich_ini(location)
        subprocess.run(["aerich", "init", "-t", "orm_config.ORM"], check=False)
        subprocess.run(["aerich", "init-db"], check=False)
    subprocess.run(["aerich", "migrate"], check=False)


@app.command()
def migrate() -> None:
    subprocess.run(["aerich", "upgrade"], check=False)


@app.command()
def shell() -> None:
    subprocess.run(["ipython"], check=False)


@app.command()
def createsuperuser(
    email: str = typer.Option(...),
    password: str = typer.Option(..., prompt=True, hide_input=True),
    models: str | None = typer.Option(None, help="Dotted path to app models module (e.g. myapp.models)"),
) -> None:
    """Create a superuser in the default models module if present."""
    # import models from current project if available
    target_model = None
    cwd = Path.cwd()
    env_models = models or os.getenv("FD_APP_MODELS")
    if env_models:
        try:
            module = __import__(env_models, fromlist=["User"])
            target_model = getattr(module, "User", None)
        except ModuleNotFoundError:
            target_model = None
    candidate_name = None
    for candidate in cwd.iterdir():
        if candidate.is_dir() and (candidate / "models.py").exists():
            module_path = f"{candidate.name}.models"
            try:
                module = __import__(module_path, fromlist=["User"])
                target_model = getattr(module, "User", None)
                if target_model is not None:
                    candidate_name = candidate.name
                    break
            except ModuleNotFoundError:
                continue
    if target_model is None:
        typer.echo("No User model found in current project.", err=True)
        raise typer.Exit(1)

    from tortoise import Tortoise, run_async  # noqa: PLC0415

    models_module = env_models or (f"{candidate_name}.models" if candidate_name else None)

    async def _create() -> None:
        try:
            await Tortoise.init(config={
                "connections": {"default": "sqlite://db.sqlite3"},
                "apps": {"models": {"models": [models_module or "app.models", "aerich.models"], "default_connection": "default"}},
            })
            pwd = CryptContext(schemes=["bcrypt"], deprecated="auto").hash(password)
            try:
                await target_model.get(email=email)
                typer.echo("User already exists", err=True)
            except DoesNotExist:
                await target_model.create(email=email, is_active=True, password=pwd, is_superuser=True)
                typer.echo("Superuser created")
        except Exception as exc:  # pragma: no cover - surfaced via tests
            typer.echo(f"Error creating superuser: {exc}", err=True)
        finally:
            with contextlib.suppress(Exception):
                await Tortoise.close_connections()

    run_async(_create())





from __future__ import annotations

from pathlib import Path


def discover_models_modules(start_dir: Path | None = None) -> list[str]:
    base = start_dir or Path.cwd()
    modules: list[str] = []
    for entry in base.iterdir():
        if entry.is_dir() and (entry / "models.py").exists():
            modules.append(f"{entry.name}.models")
    modules.append("aerich.models")
    return modules





from __future__ import annotations

from fastapi import FastAPI

from .admin import try_call_admin_hooks
from .apps import include_app_routers
from .db import init_db
from .settings import Settings


def create_app(settings: Settings | None = None) -> FastAPI:
    app_settings = settings or Settings()  # loads from env/.env
    app = FastAPI(title=app_settings.app_name)

    # Middleware by dotted path
    for dotted in app_settings.middleware:
        module_path, cls_name = dotted.rsplit(".", 1)
        module = __import__(module_path, fromlist=[cls_name])
        middleware_cls = getattr(module, cls_name)
        app.add_middleware(middleware_cls)

    # DB
    init_db(app, app_settings)

    # Include routers from installed apps
    if app_settings.installed_apps:
        include_app_routers(app, app_settings.installed_apps)

    if app_settings.admin_enabled:
        try_call_admin_hooks(app, app_settings)

    return app





from __future__ import annotations

from fastapi import FastAPI

from fast_django.settings import Settings


def init_admin(app: FastAPI, settings: Settings) -> None:
    try:
        from fastapi_admin.app import app as admin_app  # type: ignore  # noqa: PLC0415

        app.mount(settings.admin_path, admin_app)
    except Exception:
        # Fallback: mount a placeholder app to avoid hard dependency in tests
        placeholder = FastAPI()
        @placeholder.get("/")
        def _placeholder() -> dict[str, str]:
            return {"admin": "not-configured"}
        app.mount(settings.admin_path, placeholder)


